@@ -154,7 +154,7 @@ translate_device_classes (Display             *xdisplay,
                                     (XIValuatorClassInfo *) class_info);
           break;
 
-#ifdef XINPUT_2_2
+#ifdef HAVE_XINPUT_2_2
         case XIScrollClass:
           {
             XIScrollClassInfo *scroll_info = (XIScrollClassInfo *) class_info;
@@ -178,7 +178,7 @@ translate_device_classes (Display             *xdisplay,
                                                    scroll_info->increment);
           }
           break;
-#endif /* XINPUT_2_2 */
+#endif /* HAVE_XINPUT_2_2 */
 
         default:
           break;
@@ -192,7 +192,7 @@ is_touch_device (XIAnyClassInfo         **classes,
                  ClutterInputDeviceType  *device_type,
                  guint                   *n_touch_points)
 {
-#ifdef XINPUT_2_2
+#ifdef HAVE_XINPUT_2_2
   guint i;
 
   for (i = 0; i < n_classes; i++)
@@ -251,9 +251,6 @@ create_device (ClutterDeviceManagerXI2 *manager_xi2,
         source = CLUTTER_ERASER_DEVICE;
       else if (strstr (name, "cursor") != NULL)
         source = CLUTTER_CURSOR_DEVICE;
-      else if (strstr (name, "finger") != NULL ||
-               (strstr (name, "touch") != NULL && strstr (name, "touchpad") == NULL))
-        source = CLUTTER_TOUCHSCREEN_DEVICE;
       else if (strstr (name, "wacom") != NULL || strstr (name, "pen") != NULL)
         source = CLUTTER_PEN_DEVICE;
       else
@@ -427,6 +424,7 @@ translate_hierarchy_event (ClutterBackendX11       *backend_x11,
           ClutterInputDevice *master, *slave;
           XIDeviceInfo *info;
           int n_devices;
+          gboolean send_changed = FALSE;
 
           CLUTTER_NOTE (EVENT, "Hierarchy event: slave %s",
                         (ev->info[i].flags & XISlaveAttached)
@@ -442,6 +440,8 @@ translate_hierarchy_event (ClutterBackendX11       *backend_x11,
             {
               _clutter_input_device_remove_slave (master, slave);
               _clutter_input_device_set_associated_device (slave, NULL);
+
+              send_changed = TRUE;
             }
 
           /* and attach the slave to the new master if needed */
@@ -455,8 +455,18 @@ translate_hierarchy_event (ClutterBackendX11       *backend_x11,
               _clutter_input_device_set_associated_device (slave, master);
               _clutter_input_device_add_slave (master, slave);
 
+              send_changed = TRUE;
               XIFreeDeviceInfo (info);
             }
+
+          if (send_changed)
+            {
+              ClutterStage *stage = _clutter_input_device_get_stage (master);
+              if (stage != NULL)
+                _clutter_stage_x11_events_device_changed (CLUTTER_STAGE_X11 (_clutter_stage_get_window (stage)), 
+                                                          master,
+                                                          CLUTTER_DEVICE_MANAGER (manager_xi2));
+            }
         }
     }
 }
@@ -486,11 +496,11 @@ get_event_stage (ClutterEventTranslator *translator,
     case XI_ButtonPress:
     case XI_ButtonRelease:
     case XI_Motion:
-#ifdef XINPUT_2_2
+#ifdef HAVE_XINPUT_2_2
     case XI_TouchBegin:
     case XI_TouchUpdate:
     case XI_TouchEnd:
-#endif /* XINPUT_2_2 */
+#endif /* HAVE_XINPUT_2_2 */
       {
         XIDeviceEvent *xev = (XIDeviceEvent *) xi_event;
 
@@ -652,9 +662,9 @@ scroll_valuators_changed (ClutterInputDevice *device,
 
           if (direction == CLUTTER_SCROLL_UP ||
               direction == CLUTTER_SCROLL_DOWN)
-            *dx_p = delta;
-          else
             *dy_p = delta;
+          else
+            *dx_p = delta;
         }
 
       n_val += 1;
@@ -858,10 +868,10 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                                                  stage_x11,
                                                  &xev->valuators);
 
-#ifdef XINPUT_2_2
+#ifdef HAVE_XINPUT_2_2
             if (xev->flags & XIPointerEmulated)
               _clutter_event_set_pointer_emulated (event, TRUE);
-#endif /* XINPUT_2_2 */
+#endif /* HAVE_XINPUT_2_2 */
             break;
 
           default:
@@ -910,10 +920,10 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                       event->button.y,
                       event->button.axes != NULL ? "yes" : "no");
 
-#ifdef XINPUT_2_2
+#ifdef HAVE_XINPUT_2_2
         if (xev->flags & XIPointerEmulated)
           _clutter_event_set_pointer_emulated (event, TRUE);
-#endif /* XINPUT_2_2 */
+#endif /* HAVE_XINPUT_2_2 */
 
         if (xi_event->evtype == XI_ButtonPress)
           _clutter_stage_x11_set_user_time (stage_x11, event->button.time);
@@ -990,10 +1000,10 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         if (source_device != NULL && device->stage != NULL)
           _clutter_input_device_set_stage (source_device, device->stage);
 
-#ifdef XINPUT_2_2
+#ifdef HAVE_XINPUT_2_2
         if (xev->flags & XIPointerEmulated)
           _clutter_event_set_pointer_emulated (event, TRUE);
-#endif /* XINPUT_2_2 */
+#endif /* HAVE_XINPUT_2_2 */
 
         CLUTTER_NOTE (EVENT, "motion: win:0x%x device:%s (x:%.2f, y:%.2f, axes:%s)",
                       (unsigned int) stage_x11->xwin,
@@ -1006,7 +1016,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
       }
       break;
 
-#ifdef XINPUT_2_2
+#ifdef HAVE_XINPUT_2_2
     case XI_TouchBegin:
     case XI_TouchEnd:
       {
@@ -1034,7 +1044,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                                       GINT_TO_POINTER (xev->deviceid));
         clutter_event_set_device (event, device);
 
-        event->touch.axes = translate_axes (event->motion.device,
+        event->touch.axes = translate_axes (event->touch.device,
                                             event->motion.x,
                                             event->motion.y,
                                             stage_x11,
@@ -1087,7 +1097,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
                                       GINT_TO_POINTER (xev->deviceid));
         clutter_event_set_device (event, device);
 
-        event->touch.axes = translate_axes (event->motion.device,
+        event->touch.axes = translate_axes (event->touch.device,
                                             event->motion.x,
                                             event->motion.y,
                                             stage_x11,
@@ -1114,7 +1124,7 @@ clutter_device_manager_xi2_translate_event (ClutterEventTranslator *translator,
         retval = CLUTTER_TRANSLATE_QUEUE;
       }
       break;
-#endif /* XINPUT_2_2 */
+#endif /* HAVE_XINPUT_2_2 */
 
     case XI_Enter:
     case XI_Leave:
